/***** ktree.c ****************************************************
 * Description: Routines for constructing and using a keyword tree.
 * Reference: Gusfield, D. (1997). Algorithms on Strings, Trees,
 *    and sequences; Computer Science and Computational Biology.
 *    CUP, p. 52ff.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Wed Sep  8 10:20:28 2004.
 *****************************************************************/
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "eprintf.h"
#include "ktree.h"
#include "stringUtil.h"

/* prototypes private to constructKTree */
void addChild(KNode *parent, KNode *child);
KNode *findChild(KNode *parent, char c);
KTreeData *addKTreeData(KNode *node, KTreeData *ktd);
KNode *constructNode();
/*
 * Add a string to the keyword tree. 
 */
void ktree_addString(char *str, KTreeData *ktd, KNode *root){
  KNode *np, *child, *parent;
  int stringLength;
  int depth;
  depth = 0;
  np = root; 

  stringLength= strlen(str);
  parent = np;
  while((np = findChild(np,*str)) != NULL && depth < stringLength){
    str++;
    depth++;
    parent = np;
  }
  if(depth == stringLength && np != NULL){
    ktd->next = parent->data;
    parent->data->next = ktd;
  }else{
    np = parent;
    while(depth < stringLength){
      child = constructNode();
      child->c = *str++;
      child->depth = ++depth;
      addChild(parent,child);
      parent = child;
    }
    parent->data = ktd;
  }
}

/* KTreeData *addKTreeData(KNode *node, KTreeData *ktd){ */
/*   ktd->data = node->data; */
/*   return ktd; */
/* } */

void addChild(KNode *parent, KNode *child){
  KNode *np;

  if(parent->child == NULL){
    parent->child = child;
  }else{
    np = parent->child;
    while(np->sibling != NULL){
      np = np->sibling;
    }
    np->sibling = child;
  }
}

/*
 * Purpose: Find a child node of parent whose edge is labelled with c.
 */
KNode *findChild(KNode *parent, char c){
  KNode *np;

  if(parent->child == NULL){
    return NULL;
  }
  np = parent->child;
  while(np->c != c){
    if(np->sibling == NULL){
      break;
    }
      np = np->sibling;
  }
  if(np->c == c){
    return np;
  }else{
    return NULL;
  }
}

/* generate and initialize one element of KTreeData */
KTreeData *ktree_constructDataItem(){
  KTreeData *ktd;
  ktd = (KTreeData *)emalloc(sizeof(KTreeData));
  ktd->id = -1;
  ktd->pos = -1;
  return ktd;
}

KNode *ktree_constructRoot(){
  KNode *node;
  node = constructNode();
  node->c = '\0';
  return node;
}

KNode *constructNode(){
  KNode *node;
  if((node = (KNode *)malloc(sizeof(KNode))) == NULL){
    printf("ERROR [constructKTree]: memory allocation failure\n");
    assert(0);
  }
  node->depth = 0;
  node->child = NULL;
  node->sibling = NULL;
  node->fail = NULL;
  node->out = NULL;
  node->data = NULL;
  return node;
}

/*
 * Add failure and output links to tree during a breadth-first
 * traversal.
 * Reference: Gusfield, p. 60. 
 */
void ktree_preprocessTree(KNode *root){
  KNode *w, *wdash, *v, *vdash;
  KNode *child, *front, *back;
  char x;

  front = back = root;
  front->fail = NULL;
  front->out = NULL;
  while(front != NULL){
    v = front;
    vdash = v->out;
    /* add children of current node to queue */
    for(child=v->child; child != NULL; child=child->sibling){
      child->out = v;
      back->fail = child;
      back = child;
    }
    /* nullify end of queue */
    back->fail = NULL;
    /* move along one step in queue */
    front = front->fail;
    v->fail = v->out = NULL;
    if(v != NULL && vdash != NULL){
      /* set failure link */
      if(v == root){
	v->fail = v;
      }else if(vdash == root){
	v->fail = vdash;
      }else{
	wdash = NULL;
	x = v->c;
	w = vdash->fail;
	while((wdash = findChild(w,x)) == NULL && w != root)
	  w = w->fail;
	if(wdash != NULL && wdash->c == x )
	  v->fail = wdash;
	else
	  v->fail = root;
      }
    }
    /* set output link */
    if(v->fail == NULL)
      ;
    else if(v->fail->data != NULL)
      v->out = v->fail;
    else if(v->fail->out != NULL)
      v->out = v->fail->out;
  }
  root->fail = root;
}

/*
 * search text using the keyword tree rooted on root
 */
KTreeResult *ktree_search(char *text, KNode *root){
  int c, length;
  KNode *w, *wdash, *np;
  KTreeResult *resultH, *resultT, *result;
  int first = 1;

  w = root;
  c = 0;
  length = strlen(text);

  resultH = resultT = result = NULL;

  while(c < length){
    while((wdash = findChild(w,text[c])) != NULL){
      w = wdash;
      if(wdash->data != NULL){ 
	np = wdash;
	while(np != NULL){
	  result = (KTreeResult *)emalloc(sizeof(KTreeResult));
	  result->next = NULL;
	  result->data = np->data;
	  result->pos = c - np->depth + 1;
	  if(first){
	    first = 0;
	    resultH = result;
	    resultT = result;
	  }else{
	    resultT->next = result;
	    resultT = resultT->next;
	  }
	    np = np->out;
	}
      }
      c++;
    }
    if(w == root){
      c++;
    }else{
      w = w->fail;
    }
  }
  return resultH;
}


KNode *constructKeywordTree(){
  KNode *root;
  int i;
  TableEntry *tp, **hashtab;
  KTreeData *ktd;

  hashtab = getHashTable();

  root = ktree_constructRoot();
  for(i=0;i<HASHSIZE;i++){
    for(tp=hashtab[i];tp!=NULL;tp=tp->next){
      ktd = (KTreeData *)emalloc(sizeof(KTreeData));
      ktree_addString(tp->word, ktd, root);
      ktd->tableEntry = tp;
    }
  }
  return root;
}
