/***** sblast.c ***************************************************
 * Description: Simple, gapless BLAST
 * Reference: 
 * @Article{alt90:bas,
 *   author = 	 {Altschul, S. F. and Gish, W. and Miller, W. and Myers, E. W. and Lipman, D. J.},
 *   title = 	 {Basic local alignment search tool},
 *   journal = 	 {Journal of Molecular Biology},
 *   year = 	 1990,
 *   volume =	 215,
 *   pages =	 {403-410}
 * }
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Tue Nov 22 05:47:20 2005.
 *****************************************************************/
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "eprintf.h"
#include "ktree.h"
#include "interface.h"
#include "sequenceData.h"
#include "alignment.h"
#include "stringUtil.h"
#include "sblast.h"
#include "extend.h"

int main(int argc, char *argv[]){
  Args args;                   /* arguments */
  char *version;               /* program version */
  Sequence *query, *subject;   /* query & subject sequences */
  KNode *root = NULL;
  FILE *fp, *fpq, *fps;
  Sequence *patterns;
  KTreeResult *ktr;
  int i;
  SubstitutionMatrix *sm = NULL;
  char *strand = NULL;
  Msp *mspList, *lp;

  version = "0.7";
  setprogname2("sblast"); 

  args = getArgs(argc, argv);
  query = subject = NULL;
  fp = NULL;
  if(args.h == 1){
    printUsage(version);
    return 0; 
  }else if(args.e == 1){
    printUsage(version);
    return -1;
  }
  printf("# reading input data...");
  fflush(stdout);
  /* get score matrix */
  if(args.m == NULL){
    sm = constructDnaScoreMatrix(sm, args.A, args.I);
  }else{
    fp = efopen(args.m, "r");
    sm = readSubstitutionMatrix(sm, fp);
  }
  if(fp != NULL)
    fclose(fp);
  fpq = efopen(args.i,"r");
  if(args.j)
    fps = efopen(args.j,"r");
  else
    fps = stdin;
  /* get query */
  query = getNextSequence2(fpq,query);
  if(query == NULL){
    fprintf(stderr,"ERROR [%s]: no query sequence\n",progname());
    return -1;
  }
  /* ignore case */
  query->seq = strtoupper(query->seq);
  /* get subject */
  subject = getNextSequence2(fps,subject);
  if(subject == NULL){
    printf("ERROR [%s]: no query sequence\n",progname());
    return -1;
  }
  /* ignore case */
  subject->seq = strtoupper(subject->seq);
  printf("done\n");
  printf("# step1: generating word list from query...");
  fflush(stdout);
  patterns = getSubstrings(query,args.w);
  if(args.L)
    printf("\n");
  for(i=0;i<strlen(query->seq)-args.w+1;i++){
    if(args.L)
      printf("#Word %d: %s\n",i+1,patterns[i].seq);
    install(patterns[i].seq,patterns[i].qi);
  }
  /* construct keyword tree */
  root = constructKeywordTree();
  ktree_preprocessTree(root);
  if(args.L)
    printf("#done\n");
  else
    printf("done\n");
  printf("# step2: searching for exact matches of words in subject...");
  fflush(stdout);
  ktr = ktree_search(subject->seq, root);
  printf("done\n");
  printf("# step3: extending exact matches...");
  fflush(stdout);
  mspList = checkDna(ktr,query->seq,subject->seq,sm,strand,args);
  printf("done\n");
  if(args.o != NULL)
    fp = efopen(args.o,"w");
  else
    fp = stdout;
  if(mspList == NULL){
    printf("# no hits with S >= %.1f found\n",args.t);
  }else{
    fprintf(fp,"# qs\tqe\tss\tse\tscore\n");
    for(lp=mspList; lp!=NULL; lp=lp->next)
      fprintf(fp,"%d\t%d\t%d\t%d\t%.1f\n",lp->queryS+1,lp->queryE+1,lp->sbjctS+1,lp->sbjctE+1,lp->score);
  }
  return 0;
}


